/*****************************************************************************
  TIFF support for Qt 
  Copyright (c) 2001-03 by Dmitry V. Fedorov <www.dimin.net> <dima@dimin.net>
  Copyright (c) 1999 by Markus L. Noga <markus@noga.de>

  IMPLEMENTATION
  
  Programmer: Dima V. Fedorov <mailto:dima@dimin.net> <http://www.dimin.net/>
    
  History:
    02/16/2002 22:34:32 - First creation
    15/09/2003 18:49:00 - Tiled Tiff support
    16/09/2003 17:14:00 - Striped Tiff support
    24/09/2003 13:58:00 - fixed: small bug in paletted images
    16/10/2003 18:54:00 - fixed: PHOTOMETRIC_MINISWHITE images
    21/10/2003 14:22:00 - Resolution extraction by castle <www.castlejohn.org>
        
  Ver : 4
*****************************************************************************/
//
// Read TIFF image files with Qt
// Copyright (c) 1999 by Markus L. Noga <markus@noga.de>
//  
// Reader improvement to read 8 and 24 bits by Dmitry Fedorov
// Write TIFF image files with Qt 
// Copyright (c) 2001-2002 by Dmitry V. Fedorov <www.dimin.net>
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 

#include <stdio.h>
#include <stdlib.h>
#include <qglobal.h>

#include <qmessagebox.h>

#if defined (_WS_WIN_) || defined(Q_WS_WIN)
#define MAP_FAILED 1
#include "src/tiffio.h"
#else
#include <sys/mman.h>
#include <tiffio.h>
#endif

#include <qimage.h>
#include <qfile.h>

#include "qtiffio.h"

//! QIODevice / libtiff read wrapper
//
static tsize_t tiff_read(thandle_t handle,tdata_t data,tsize_t size) 
{
  QIODevice *iod=(QIODevice*) handle;
  return (tsize_t) iod->readBlock((char*) data,size);
}

//! QIODevice / libtiff write wrapper
//
static tsize_t tiff_write(thandle_t handle,tdata_t data,tsize_t size) 
{
  QIODevice *iod=(QIODevice*) handle;
  return (tsize_t) iod->writeBlock((const char*) data,size);
  //iod->flush();
}

//! QIODevice / libtiff seek wrapper
/*! \returns the current file position. libtiff wants that.
*/
static toff_t tiff_seek(thandle_t handle, toff_t offset, int whence) 
{
  QIODevice *iod=(QIODevice*) handle;
  
  if(whence==SEEK_SET)
    iod->at(offset);
  else if(whence==SEEK_CUR)
    iod->at(iod->at()+offset);
  else if(whence==SEEK_END)
    iod->at(iod->size()+offset);
  else
    return 0; // -1 dima
  
  return iod->at();
}

//! QIODevice / libtiff close wrapper
/*! This is a dummy, the IO device's owner will close it.
*/
static int tiff_close(thandle_t handle) {
  QIODevice *iod=(QIODevice*) handle;
  iod->flush();
  iod->close();
  return 0;
}

//! QIODevice / libtiff size wrapper
//
static toff_t tiff_size(thandle_t handle) {
  QIODevice *iod=(QIODevice*) handle;
  return iod->size();
}

//! QIODevice / libtiff mmap wrapper
/*! \warning always returns MAP_FAILED.
*/
static int tiff_mmap(thandle_t handle,tdata_t* data,toff_t* size) {
  handle=handle; data=data; size=size;
  return (int) MAP_FAILED;
}

//! QIODevice / libtiff write wrapper
/*! \warning because you can't mmap, this is a dummy.
*/
static void tiff_unmap(thandle_t handle, tdata_t data, toff_t size) {
  handle=handle; data=data; size=size;
}

static bool is_straight_copy()
{
  if ( TIFFGetR(0x1234567) == qRed   (0x1234567) &&
       TIFFGetG(0x1234567) == qGreen (0x1234567) &&
       TIFFGetB(0x1234567) == qBlue  (0x1234567) ) 
  return TRUE;
  else 
  return FALSE;
} 

static void init_qimage_palette(QImage *img, TIFF *tif)
{
  int i;
  QColor color;
  uint16 photometric = PHOTOMETRIC_MINISWHITE;
  uint16 bitspersample = 1;

  if (tif == NULL) return;
  if (img == NULL) return;
  
  TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric);
  TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample);
  
  if (photometric == PHOTOMETRIC_PALETTE)
  { // palette
    uint16 *red;
    uint16 *green;
    uint16 *blue;
    
    TIFFGetField(tif, TIFFTAG_COLORMAP, &red, &green, &blue);
    
    for (i=(1 << bitspersample)-1; i>=0; i--)
    {
      if ( (red[i] > 255) || (green[i] > 255) || (blue[i] > 255) )
        color.setRgb(i, i, i);  
      else
        color.setRgb(red[i], green[i], blue[i]);
    
      img->setColor(i, color.rgb());
    }
  }
  else
  { 
    // if greyscale images
    //if (photometric == PHOTOMETRIC_MINISBLACK) 
    for (i=0; i<256; i++)
    {
      color.setRgb(i, i, i);
      img->setColor(i, color.rgb());
    }
  }
  
}

//****************************************************************************
// TILED TIFF
//****************************************************************************
static void read_tiled_RGBA_tiff(TIFF *tif, QImage *img)
{
  if (tif == NULL) return;
  if (img == NULL) return;

  uint32 columns, rows;
  uint32 *tile_buf;
  uint32 x, y;

  // if tiff is not tiled get out and never come back :-)
  if( !TIFFIsTiled(tif) ) return;

  TIFFGetField(tif, TIFFTAG_TILEWIDTH,  &columns);
  TIFFGetField(tif, TIFFTAG_TILELENGTH, &rows);

  tile_buf = (uint32*) _TIFFmalloc(columns*rows*sizeof(uint32));
  if (tile_buf == NULL) return;

  uint32 tileW = columns, tileH = rows;

  for (y = 0; y < (uint32)img->height(); y += rows)
  {
    if (y > (uint)img->height()) break;

    // the tile height may vary 
    if (img->height()-y < rows) tileH = img->height()-y; 

    tileW = columns;
    for (x = 0; x < (uint)img->width(); x += columns)
    {

      // eventually i culd do this test, it's much faster
      //if( TIFFReadEncodedTile( poGDS->hTIFF, nBlockId, pImage, nBlockBufSize ) == -1 )  

      if (!TIFFReadRGBATile(tif, x, y, tile_buf)) break;

      // the tile size is now treated by libtiff guys the
      // way that the size stay on unchanged      
      //if (img->width()-x < columns) tileW = img->width()-x;

      // now put tile into the image
      if (is_straight_copy() == TRUE)
      {
        register uint32 yi;          
        // just mirror on x axis.
        for(yi = 0; yi < tileH; yi++) 
          _TIFFmemcpy(img->scanLine(y+(tileH-1-yi))+x, tile_buf+tileW*yi, tileW*4);
      } 
      else 
      {
        uint32 *inp=tile_buf;
        register uint32 xi, yi;

        //for(yi = tileH; yi > 0; yi--) 
        for(yi = 0; yi < tileH; yi++) 
        {
          QRgb *row= (QRgb*) ( img->scanLine(y+yi) );
          for(xi = 0; xi < tileW; xi++)  
          {
            const uint32 pix = *(tile_buf + ((tileH-1-yi)*tileW)+xi);
            row[xi + x] = qRgb(TIFFGetR(pix), 
                               TIFFGetG(pix), 
                               TIFFGetB(pix) ) | (TIFFGetA(pix)<<24);
          } // for xi
        } // for yi
      } // else per pixel method
    } // for x by tiles
  } // for y by tiles

  _TIFFfree(tile_buf);
}

static void read_tiled_8BIT_tiff(TIFF *tif, QImage *img)
{
  if (tif == NULL) return;
  if (img == NULL) return;

  uint32 columns, rows;
  uchar *tile_buf;
  uint32 x, y;

  // if tiff is not tiled get out and never come back :-)
  if( !TIFFIsTiled(tif) ) return;

  TIFFGetField(tif, TIFFTAG_TILEWIDTH,  &columns);
  TIFFGetField(tif, TIFFTAG_TILELENGTH, &rows);

  tile_buf = (uchar*) _TIFFmalloc(TIFFTileSize(tif));

  if (tile_buf == NULL) return;

  uint32 tileW = columns, tileH = rows;

  for (y = 0; y < (uint32)img->height(); y += rows)
  {
    if (y > (uint)img->height()) break;

    // the tile height may vary 
    if (img->height()-y < rows) tileH = img->height()-y; 

    tileW = columns;
    for (x = 0; x < (uint)img->width(); x += columns)
    {
      register uint32 yi;
      uint32 tW;
      
      /*
      ttile_t tile_id;
      tile_id = TIFFComputeTile(tif, x, y, 0, 0);
      if (!TIFFReadEncodedTile(tif, tile_id, tile_buf, TIFFTileSize(tif))) break;
      */
      
      if (!TIFFReadTile(tif, tile_buf, x, y, 0, 0)) break;

      // the tile size is now treated by libtiff guys the
      // way that the size stay on unchanged      
      if (img->width()-x < columns) tW = img->width()-x; else tW = tileW;

      // now put tile into the image
      for(yi = 0; yi < tileH; yi++) 
      {
        uchar *row= (uchar*) ( img->scanLine(y+yi) );
        _TIFFmemcpy(row+x, tile_buf+yi*tileW, tW);
      }
    } // for x
  } // for y

  _TIFFfree(tile_buf);
}

//****************************************************************************
// STRIPED TIFF
//****************************************************************************
static int read_striped_RGBA_tiff(TIFF *tif, QImage *img)
{
  int result = -1;
  if (tif == NULL) return result;
  if (img == NULL) return result;

  uint32 width, rows;
  uint32 *buf=NULL;
  uint32 y;

  // check if tiff is striped
  if( !TIFFGetField(tif, TIFFTAG_ROWSPERSTRIP, &rows) ) return result;
  TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);

  buf = (uint32*) _TIFFmalloc(width*rows*sizeof(uint32));
  if (buf == NULL) return result;

  uint32 tileH = rows;

  for (y = 0; y < (uint32)img->height(); y += rows)
  {
    if (y > (uint)img->height()) break;

    // the tile height may vary 
    if (img->height()-y < rows) tileH = img->height()-y; 

    if (!TIFFReadRGBAStrip(tif, y, buf)) break;

    // now put strip into the image
    if (is_straight_copy() == TRUE)
    {
      uint32 yi;
      for(yi = 0; yi < tileH; yi++) 
        _TIFFmemcpy(img->scanLine(y+(tileH-1-yi)), buf+yi*width, width*4);
    } // straight copy
    else 
    { // not straight copy
      uint32 xi, yi;
      uint32 *inp = buf;

      for(yi = 0; yi < tileH; yi++) 
      {
        QRgb *row= (QRgb*) ( img->scanLine(y+(tileH-1-yi)) );
        for(xi = 0; xi < width; xi++)  
        {
          const uint32 col = *(inp++);
          row[xi] = qRgba(TIFFGetR(col), TIFFGetG(col), TIFFGetB(col), TIFFGetA(col) );
         } // for xi
      } // for yi
    } // not straight copy
    
    // read Alpha channel
 

    result = 0;
  } // for y

  _TIFFfree(buf);

  return result;
}

//****************************************************************************
// SCANLINE METHOD TIFF
//****************************************************************************

// the simplies default method, it takes too many memory to load
static int read_scanline_8BIT_tiff(TIFF *tif, QImage *img)
{
  int result = -1;
  register uint32 y = 0;
  if (tif == NULL) return result;
  if (img == NULL) return result;
  uint32 height = 0; 
  TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height);

  for(y=0; y<height; y++) {
    result = TIFFReadScanline(tif, img->scanLine(y), y, 0);
  }
  return result;
}


//****************************************************************************
// DEFAULT METHOD TIFF
//****************************************************************************

// the simplies default method, it takes too many memory to load
static void read_generic_RGBA_tiff(TIFF *tif, QImage *img)
{
  if (tif == NULL) return;
  if (img == NULL) return;

  unsigned size = 0;
  uint32 height = 0; 
  uint32 width = 0; 

  TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
  TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height);
  size=width*height;

  uint32 *bits=(uint32*) _TIFFmalloc(size * sizeof (uint32));
  if (!bits) return;

  if (TIFFReadRGBAImage(tif, width, height, bits, 0)) 
  {
    if (is_straight_copy() == TRUE)
    {
      // just mirror on x axis.
      for(unsigned y=0; y<height; y++) memcpy(img->scanLine(height-1-y),bits+y*width,width*4);
    } 
    else 
    {
      // swap bytes
      uint32 *inp=bits;
      for(unsigned y=0; y<height; y++) 
      {
        QRgb *row=(QRgb*) (img->scanLine(height-1-y));
        for(unsigned x=0; x<width; x++) 
        {
          const uint32 col=*(inp++);
          row[x]=qRgb(TIFFGetR(col), TIFFGetG(col), TIFFGetB(col) ) | (TIFFGetA(col)<<24);
        }
      }
    } // else 
  } // if read RGBA
  
  _TIFFfree(bits);
}

//! QImageIO read handler for TIFF files.
//
static void read_tiff_image(QImageIO *iio) {
  QImage img;
  int state=-1;
  
  uint32 height = 0; 
  uint32 width = 0; 
  uint16 bitspersample = 1;
  uint16 samplesperpixel = 1;
  uint32 rowsperstrip;  
  uint16 photometric = PHOTOMETRIC_MINISWHITE;
  uint16 compression = COMPRESSION_NONE;
  uint16 PlanarConfig;
  
  // resolution support by castle
  float xresolution = 0;
  float yresolution = 0;
  short resolutionunit = 0;

  
  bool isimg;
  
  // pass on file name if known
  const char *fileName;
  QFile *f = (QFile*)(iio->ioDevice());
  
  if (f != 0) fileName = f->name();
  else fileName = "QIODevice";
  
  // open without memory mapping.
  TIFF *tif=TIFFClientOpen(fileName,"rm",
    (thandle_t) (iio->ioDevice()),
    tiff_read,
    tiff_write,
    tiff_seek,
    tiff_close,
    tiff_size,
    tiff_mmap,
    tiff_unmap );
  
  if(tif) 
  {
    TIFFGetField(tif, TIFFTAG_COMPRESSION, &compression);
    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height);
    TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);
    TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample);
    TIFFGetField(tif, TIFFTAG_ROWSPERSTRIP, &rowsperstrip);   
    TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric);
    
    TIFFGetField(tif, TIFFTAG_PLANARCONFIG, &PlanarConfig);	// single image plane 

    if ( (bitspersample == 8) && (samplesperpixel == 1) )
      isimg = img.create(width, height, 8, 256, QImage::IgnoreEndian);
    else   
      isimg = img.create(width, height, 32);

    if (isimg == TRUE)
    {
      //-----------------------------------------------------------------------
      // image resolution setting - by castle
      //-----------------------------------------------------------------------      
      if (
        ( TIFFGetField(tif, TIFFTAG_RESOLUTIONUNIT , &resolutionunit) )
        &&
        ( TIFFGetField(tif, TIFFTAG_XRESOLUTION , &xresolution) )
        &&
        ( TIFFGetField(tif, TIFFTAG_YRESOLUTION , &yresolution) )
        )
      {
        if (resolutionunit == RESUNIT_INCH)
        {
          int x_m ;
          int y_m ;
          x_m = (int)(xresolution*0.0254) ;
          y_m = (int)(yresolution*0.0254) ;
          img.setDotsPerMeterX( x_m );
          img.setDotsPerMeterY( y_m );
        }
        else if (resolutionunit == RESUNIT_CENTIMETER)
        {
          int x_m ;
          int y_m ;
          x_m = (int)(xresolution/100) ;
          y_m = (int)(yresolution/100) ;
          img.setDotsPerMeterX( x_m );
          img.setDotsPerMeterY( y_m );
        };
      };
      //-----------------------------------------------------------------------
      
      if ( (bitspersample == 8) && (samplesperpixel == 1) )
      { // paletted image
        init_qimage_palette(&img, tif);
      
        if( !TIFFIsTiled(tif) )
          read_scanline_8BIT_tiff(tif, &img);
        else
          read_tiled_8BIT_tiff(tif, &img);

        // invert each pixel if PHOTOMETRIC_MINISWHITE
        if (photometric == PHOTOMETRIC_MINISWHITE)
        {
          ulong size = img.width() * img.height();
          uchar *px = img.bits();
          ulong p;
          for (p=0; p<size; p++) *(px + p) = 255 - *(px + p);
        } // if PHOTOMETRIC_MINISWHITE

      } // end paletted
      else
      { // RGBA or anything else readed as RGBA
        if( !TIFFIsTiled(tif) )
        {
          if (read_striped_RGBA_tiff(tif, &img) == -1) 
            read_generic_RGBA_tiff(tif, &img); 
        }
        else
          read_tiled_RGBA_tiff(tif, &img);
      } // RGBA

      iio->setImage(img);
      state = 0;
    } // if image creation is ok       
    TIFFClose(tif);
  } // if tif
  iio->setStatus(state);
}

//! QImageIO write handler for TIFF files.
//
static void write_tiff_image(QImageIO *iio) 
{
  QImage img;
  int state=-1;
  
  const char *fileName;
  QFile *f = (QFile*)(iio->ioDevice());
  if (f != 0) fileName = f->name();
  else fileName = "QIODevice";
  img = iio->image();
  
  iio->setStatus(state);
  
  TIFF *out=TIFFClientOpen(fileName,"w",
    (thandle_t) (iio->ioDevice()),
    tiff_read,
    tiff_write,
    tiff_seek,
    tiff_close,
    tiff_size,
    tiff_mmap,
    tiff_unmap );
  
  if(out) 
  {
    uint32 height;
    uint32 width;
    uint32 rowsperstrip = (uint32) -1;
    uint16 bitspersample;
    uint16 samplesperpixel;
    uint16 photometric = PHOTOMETRIC_RGB;
    uint16 compression;
    
    uint32 x, y;
    
    width = img.width();
    height = img.height();
    bitspersample = img.depth();
    samplesperpixel = ((bitspersample == 24) || (bitspersample == 32)) ? 3 : 1;
    
    if (img.depth()>=24) photometric = PHOTOMETRIC_RGB;
    if ((img.depth()<24)&&(img.isGrayscale()!=TRUE)) photometric = PHOTOMETRIC_PALETTE;
    if ((img.depth()<24)&&(img.isGrayscale()==TRUE)) photometric = PHOTOMETRIC_MINISBLACK;
    
    // handle standard width/height/bpp stuff
    TIFFSetField(out, TIFFTAG_IMAGEWIDTH, width);
    TIFFSetField(out, TIFFTAG_IMAGELENGTH, height);
    TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, samplesperpixel);
    TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, ((bitspersample == 32) ? 24 : bitspersample) / samplesperpixel);
    TIFFSetField(out, TIFFTAG_PHOTOMETRIC, photometric);
    TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);	// single image plane 
    TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, TIFFDefaultStripSize(out, rowsperstrip));

    TIFFSetField(out, TIFFTAG_SOFTWARE, "DIMIN QT TIFF WRAPPER <www.dimin.net>");
    /*
    if( TIFFGetField( hTIFF, TIFFTAG_DOCUMENTNAME, &pszText ) )
    if( TIFFGetField( hTIFF, TIFFTAG_IMAGEDESCRIPTION, &pszText ) )
    if( TIFFGetField( hTIFF, TIFFTAG_DATETIME, &pszText ) )
    */
    
    // handle metrics
    //TIFFSetField(out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    //TIFFSetField(out, TIFFTAG_XRESOLUTION, 0.0254*30.0);
    //TIFFSetField(out, TIFFTAG_YRESOLUTION, 0.0254*30.0);
    
    // only one page
    TIFFSetField(out, TIFFTAG_SUBFILETYPE, 0);
    
    TIFFFlushData(out);
    
    // palettes (image colormaps are automatically scaled to 16-bits)
    if (photometric == PHOTOMETRIC_PALETTE) {
      uint16 *r, *g, *b;
      uint16 nColors = img.numColors();
      
      r = (uint16 *) _TIFFmalloc(sizeof(uint16) * 3 * nColors);
      g = r + nColors;
      b = g + nColors;
      
      for (int i = nColors - 1; i >= 0; i--) 
      {
        r[i] = (uint16) qRed(img.color(i));
        g[i] = (uint16) qGreen(img.color(i));
        b[i] = (uint16) qBlue(img.color(i));
      }
      
      TIFFSetField(out, TIFFTAG_COLORMAP, r, g, b);
      
      _TIFFfree(r);
    }
    
    f->flush();
    TIFFFlushData(out);
    
    // compression
    switch(bitspersample) {
    case 1 :
      compression = COMPRESSION_CCITTFAX4;
      break;
      
    case 8 :
    case 24 :
    case 32 :
      //compression = COMPRESSION_PACKBITS;
      compression = COMPRESSION_NONE;
      break;
      
    default :
      compression = COMPRESSION_NONE;
      break;
    }
    
    TIFFSetField(out, TIFFTAG_COMPRESSION, compression);
    f->flush();
    TIFFFlushData(out);
    
    // read the DIB lines from bottom to top
    // and save them in the TIF
    // -------------------------------------
    
    switch(bitspersample) {				
    case 1 :
    case 4 :
    case 8 :
      {
        for (y = 0; y < height; y++) {
          BYTE *bits = img.scanLine(y);
          TIFFWriteScanline(out,bits, y, 0);
          f->flush();
          TIFFFlushData(out);
        }
        break;
      }				
      
    case 24:
    case 32:
      {
        BYTE *buffer = (BYTE *)calloc(width,3);
        QRgb *bufIn;
        
        for (y = 0; y < height; y++) 
        {
          // TIFFs store color data RGB instead of BGR
          BYTE *pBuf = buffer;
          bufIn = (QRgb*) img.scanLine(y);
          for (x=0; x<width; x++) 
          {
            pBuf[0] = (BYTE) qRed(*(bufIn+x));
            pBuf[1] = (BYTE) qGreen(*(bufIn+x));
            pBuf[2] = (BYTE) qBlue(*(bufIn+x));
            pBuf += 3;
          }
          
          // write the scanline to disc
          TIFFWriteScanline(out, buffer, y, 0);
          f->flush();
          TIFFFlushData(out);
        }
        free(buffer);
        break;
      }
    }
    TIFFFlushData(out);
    f->flush();
    TIFFClose(out);
    state=0;
  }
  
  iio->setStatus(state);
}

//! Add TIFF format capability to QImage.
//
void qInitTiffIO() {
  // Qt regexps don't cut it - we're missing an OR-Operator.
  QImageIO::defineIOHandler("TIFF", "^[MI][MI][\\x01*][\\x01*]", 0, read_tiff_image, write_tiff_image);
}
